@@ -3,7 +3,7 @@
  *
  * pdfresurrect - PDF history extraction tool
  *
- * Copyright (C) 2008-2010, 2012-2013, 2017-18, Matt Davis (enferex).
+ * Copyright (C) 2008-2010, 2012-2013, 2017-19, Matt Davis (enferex).
  *
  * Special thanks to all of the contributors:  See AUTHORS.
  *
@@ -122,7 +122,7 @@ pdf_t *pdf_new(const char *name)
     const char *n;
     pdf_t      *pdf;
    
-    pdf = calloc(1, sizeof(pdf_t));
+    pdf = safe_calloc(sizeof(pdf_t));
 
     if (name)
     {
@@ -132,12 +132,12 @@ pdf_t *pdf_new(const char *name)
         else
           n = name;
 
-        pdf->name = malloc(strlen(n) + 1);
+        pdf->name = safe_calloc(strlen(n) + 1);
         strcpy(pdf->name, n);
     }
     else /* !name */
     {
-        pdf->name = malloc(strlen("Unknown") + 1);
+        pdf->name = safe_calloc(strlen("Unknown") + 1);
         strcpy(pdf->name, "Unknown");
     }
 
@@ -216,7 +216,7 @@ int pdf_load_xrefs(FILE *fp, pdf_t *pdf)
 
     /* Load in the start/end positions */
     fseek(fp, 0, SEEK_SET);
-    pdf->xrefs = calloc(1, sizeof(xref_t) * pdf->n_xrefs);
+    pdf->xrefs = safe_calloc(sizeof(xref_t) * pdf->n_xrefs);
     ver = 1;
     for (i=0; i<pdf->n_xrefs; i++)
     {
@@ -233,6 +233,11 @@ int pdf_load_xrefs(FILE *fp, pdf_t *pdf)
           fseek(fp, pos - (++pos_count), SEEK_SET);
         
         /* Suck in end of "startxref" to start of %%EOF */
+        if (pos_count >= sizeof(buf)) {
+          ERR("Failed to locate the startxref token. "
+              "This might be a corrupt PDF.\n");
+          return -1;
+        }
         memset(buf, 0, sizeof(buf));
         SAFE_E(fread(buf, 1, pos_count, fp), pos_count,
                "Failed to read startxref.\n");
@@ -309,7 +314,7 @@ void pdf_load_pages_kids(FILE *fp, pdf_t *pdf)
 
             /* Get root catalog */
             sz = pdf->xrefs[i].end - ftell(fp);
-            buf = malloc(sz + 1);
+            buf = safe_calloc(sz + 1);
             SAFE_E(fread(buf, 1, sz, fp), sz, "Failed to load /Root.\n");
             buf[sz] = '\0';
             if (!(c = strstr(buf, "/Root")))
@@ -439,7 +444,7 @@ void pdf_summarize(
 
     if (name)
     {
-        dst_name = malloc(strlen(name) * 2 + 16);
+        dst_name = safe_calloc(strlen(name) * 2 + 16);
         sprintf(dst_name, "%s/%s", name, name);
 
         if ((c = strrchr(dst_name, '.')) && (strncmp(c, ".pdf", 4) == 0))
@@ -599,10 +604,10 @@ static int is_valid_xref(FILE *fp, pdf_t *pdf, xref_t *xref)
 
         if (c && xref->is_stream)
         {
-            free(c);
             pdf->has_xref_streams = 1;
             is_valid = 1;
         }
+        free(c);
     }
 
     fseek(fp, start, SEEK_SET);
@@ -622,7 +627,7 @@ static void load_xref_entries(FILE *fp, xref_t *xref)
 static void load_xref_from_plaintext(FILE *fp, xref_t *xref)
 {
     int  i, buf_idx, obj_id, added_entries;
-    char c, buf[21];
+    char c, buf[32] = {0};
     long start, pos;
 
     start = ftell(fp);
@@ -638,7 +643,7 @@ static void load_xref_from_plaintext(FILE *fp, xref_t *xref)
 
     SAFE_E(fread(buf, 1, 21, fp), 21, "Failed to load entry Size string.\n");
     xref->n_entries = atoi(buf + strlen("ize "));
-    xref->entries = calloc(1, xref->n_entries * sizeof(struct _xref_entry));
+    xref->entries = safe_calloc(xref->n_entries * sizeof(struct _xref_entry));
 
     /* Load entry data */
     obj_id = 0;
@@ -646,17 +651,25 @@ static void load_xref_from_plaintext(FILE *fp, xref_t *xref)
     added_entries = 0;
     for (i=0; i<xref->n_entries; i++)
     {
-        /* Skip */
+        /* Advance past newlines. */
         c = fgetc(fp);
         while (c == '\n' || c == '\r')
           c = fgetc(fp);
 
+        /* Collect data up until the following newline. */
         buf_idx = 0;
-        while (c != '\n' && c != '\r')
+        while (c != '\n' && c != '\r' && !feof(fp) &&
+               !ferror(fp) && buf_idx < sizeof(buf))
         {
             buf[buf_idx++] = c;
             c = fgetc(fp);
         }
+        if (buf_idx >= sizeof(buf))
+        {
+            ERR("Failed to locate newline character. "
+                "This might be a corrupt PDF.\n");
+            exit(EXIT_FAILURE);
+        }
         buf[buf_idx] = '\0';
 
         /* Went to far and hit start of trailer */
@@ -797,7 +810,7 @@ static pdf_creator_t *new_creator(int *n_elements)
         {"Trapped",      ""},
     };
 
-    daddy = malloc(sizeof(creator_template));
+    daddy = safe_calloc(sizeof(creator_template));
     memcpy(daddy, creator_template, sizeof(creator_template));
 
     if (n_elements)
@@ -818,7 +831,7 @@ static pdf_creator_t *new_creator(int *n_elements)
 static void load_creator(FILE *fp, pdf_t *pdf)
 {
     int    i, buf_idx;
-    char   c, *buf, obj_id_buf[32];
+    char   c, *buf, obj_id_buf[32] = {0};
     long   start;
     size_t sz;
 
@@ -848,7 +861,7 @@ static void load_creator(FILE *fp, pdf_t *pdf)
         while (SAFE_F(fp, (!isspace(c = fgetc(fp)) && (c != '>'))))
             ; /* Iterate to first white space /Info<space><data> */
 
-        /* No space between /Info and it's data */
+        /* No space between /Info and its data */
         END_OF_TRAILER(c, start, fp);
 
         while (SAFE_F(fp, (isspace(c = fgetc(fp)) && (c != '>'))))
@@ -860,12 +873,11 @@ static void load_creator(FILE *fp, pdf_t *pdf)
         /* Get obj id as number */
         buf_idx = 0;
         obj_id_buf[buf_idx++] = c;
-        while (SAFE_F(fp, (!isspace(c = fgetc(fp)) && (c != '>'))))
+        while ((buf_idx < (sizeof(obj_id_buf) - 1)) &&
+               SAFE_F(fp, (!isspace(c = fgetc(fp)) && (c != '>'))))
           obj_id_buf[buf_idx++] = c;
 
         END_OF_TRAILER(c, start, fp);
-
-        obj_id_buf[buf_idx] = '\0';
      
         /* Get the object for the creator data.  If linear, try both xrefs */ 
         buf = get_object(fp, atoll(obj_id_buf), &pdf->xrefs[i], &sz, NULL);
@@ -1083,12 +1095,11 @@ static char *get_object(
     /* Jump to object start */
     fseek(fp, entry->offset, SEEK_SET);
 
-    /* Initial allocate */
+    /* Initial allocation */
     obj_sz = 0;    /* Bytes in object */
     total_sz = 0;  /* Bytes read in   */
     n_blks = 1;
-    data = malloc(blk_sz * n_blks);
-    memset(data, 0, blk_sz * n_blks);
+    data = safe_calloc(blk_sz * n_blks);
 
     /* Suck in data */
     stream = 0;
@@ -1100,6 +1111,10 @@ static char *get_object(
 
         if (total_sz + blk_sz >= (blk_sz * n_blks))
           data = realloc(data, blk_sz * (++n_blks));
+        if (!data) {
+          ERR("Failed to reallocate buffer.\n");
+          exit(EXIT_FAILURE);
+        }
 
         search = total_sz - read_sz;
         if (search < 0)
@@ -1254,18 +1269,12 @@ static int get_page(int obj_id, const xref_t *xref)
 
 static char *get_header(FILE *fp)
 {
-    long start;
-
     /* First 1024 bytes of doc must be header (1.7 spec pg 1102) */
-    char *header;
-
-    header = calloc(1, 1024);
-    
-    start = ftell(fp);
+    char *header = safe_calloc(1024);
+    long start = ftell(fp);
     fseek(fp, 0, SEEK_SET);
     SAFE_E(fread(header, 1, 1023, fp), 1023, "Failed to load PDF header.\n");
     fseek(fp, start, SEEK_SET);
-    
     return header;
 }
 
@@ -1280,7 +1289,7 @@ static char *decode_text_string(const char *str, size_t str_len)
     /* Regular encoding */
     if (str[0] == '(')
     {
-        ascii = malloc(strlen(str) + 1);
+        ascii = safe_calloc(strlen(str) + 1);
         strncpy(ascii, str, strlen(str) + 1);
         return ascii;
     }
@@ -1302,7 +1311,7 @@ static char *decode_text_string(const char *str, size_t str_len)
       return NULL;
 
     /* Now decode as hex */
-    ascii = malloc(str_len);
+    ascii = safe_calloc(str_len);
     for ( ; idx<str_len; ++idx)
     {
         hex_buf[0] = str[idx++];
